#                                                         -*- Perl -*-
# wdic-menu - FreePWING script for WDIC 2001
# $Id: wdic-menu.pl 1.2.2.0 2001/07/25 14:55:03 satomii Exp $
#
# This file is designed to be used from the wdic-fpw.pl script.
#
# KANJI-CODE: SJIS; TAB-SIZE: 4
# DO NOT CHANGE THE KANJI CODE OF THIS SCRIPT. IT MUST BE SHIFT JIS.
#
# Copyright (C) 2000-2001, rei@wdic.org.
# This program is distributed in the hope that it will be useful, but
# without any warranty. See the GNU General Public License for the details.
#
require 5.005;

use File::DosGlob 'glob';
use strict;

my @copyright;
my $index_title = 'O[vʎ^ꗗ';

#-----------------------------------------------------------------------#
#	menu construction													#
#-----------------------------------------------------------------------#
#
# void make_menu(void)
# writes the description about each groups in the current text context.
#
sub make_menu()
{
	my(@handle, @list, $title, $index, $str, $prev_toc, $next_toc);
	my $menu_tag = 'WDIC_MAN_%02X';

	print "Writing menu and copyright pages...\n";

	#
	# first create the top level menu.
	#
	# readme.txt file, required.
	@list = glob("./readme.txt");
	if (@list) {
		$handle[0] = new IO::File($list[0], 'r');
	} else {
		die "Unable to open file './readme.txt' ($!).\n";
	}
	binmode $handle[0];
	while ($_ = $handle[0]->getline()) {
		if (s/^([^\r\n]+[\r\n]+)$/$1/) {
			&write_link($fpwmenu, $_, 'README');
			last;
		}
	}

	# WDIC.MAN file, required.
	@list = glob("$wdic_dir/WDIC.MAN");
	if (@list) {
		$handle[1] = new IO::File($list[0], 'r');
	} else {
		die "Unable to open file '$wdic_dir/WDIC.MAN' ($!).\n";
	}
	binmode $handle[1];
	while ($_ = $handle[1]->getline()) {
		if (s/^\s*(ʐMp̊bm)\s*([0-9]+[\r\n]+)$/$1 $2/) {
			&write_link($fpwmenu, $_, 'WDIC_MAN');
			$title = $_;
			last;
		}
	}

	# QUESTION.TXT file, optional.
	@list = glob("$wdic_dir/QUESTION.TXT");
	if (@list) {
		$handle[2] = new IO::File($list[0], 'r');
		if (defined($handle[2])) {
			binmode $handle[2];
			while ($_ = $handle[2]->getline()) {
				if (s/^\s+([^\r\n]+[\r\n]+)$/$1/) {
					&write_link($fpwmenu, $_, 'QUESTION');
					last;
				}
			}
		}
	}

	# index by group, if tag files are found.
	@list = glob("$wdic_dir/*.TAG");
	&write_link($fpwmenu, "$index_title\n", 'INDEX') if (@list);

	#
	# write the body of readme.txt file.
	#
	print "readme.txt\n";
	$handle[0]->seek(0, IO::File::SEEK_SET);
	while ($_ = $handle[0]->getline()) {
		last if (s/^([^\r\n]+)[\r\n]+$/$1/);
	}
	&new_menu_context($_, 'README');
	while ($_ = $handle[0]->getline()) {
		&write_sjis($fpwmenu, $_);
	}
	$handle[0]->close();

	#
	# write the body of WDIC.MAN file.
	#
	print "WDIC.MAN\n";
	&new_menu_context('', 'WDIC_MAN');

	$index = 0;
	while ($_ = $handle[1]->getline()) {
		if (s/^\-+  (.+)  \-+[\r\n]+$/$1/) {
			$index++;
			$_ = &reform_sjis($_, 1) if (/\x81[\x69\x6a]/);
			&write_link($fpwmenu, "$_\n", sprintf($menu_tag, $index));
		}
	}

	$index = 0;
	$handle[1]->seek(0, IO::File::SEEK_SET);
	while ($str = $handle[1]->getline()) {
		last if ($str =~ s/^\-+  (.+)  \-+[\r\n]+$/$1/);
	}

	do {
		print "  $str\n";
		$index++;
		&new_menu_context($str, sprintf($menu_tag, $index));
		$next_toc = &write_menu_section($str, $handle[1]);

		# menu footer (navigation).
		if (!$fpwmenu->add_indent_level(2) || !$fpwmenu->add_newline()) {
			die &death_cry("Failed to write menu footer", $fpwmenu);
		}
		if ($prev_toc) {
			&write_sjis($fpwmenu, '');
			&write_link($fpwmenu, $prev_toc, sprintf($menu_tag, $index - 1));
		}
		if ($next_toc) {
			&write_sjis($fpwmenu, ($prev_toc ? ', ' : '') . '');
			&write_link($fpwmenu, $next_toc, sprintf($menu_tag, $index + 1));
		}
		&write_sjis($fpwmenu, ', ');
		&write_link($fpwmenu, "$title\n", 'WDIC_MAN');
		$prev_toc = $str;
		$str = $next_toc;
	} while ($_);

	$handle[1]->close();

	#
	# write the body of QUESTION.TXT file if any.
	#
	if (defined($handle[2])) {
		print "QUESTION.TXT\n";
		$handle[2]->seek(0, IO::File::SEEK_SET);
		while ($_ = $handle[2]->getline()) {
			last if (s/^\s+([^\r\n]+)[\r\n]+$/$1/);
		}
		&new_menu_context($_, 'QUESTION');
		while ($_ = $handle[2]->getline()) {
			if (/^\- \- \-/) {
				&write_sjis($fpwmenu, $_);
				last;
			}
			s/^ {5}//;
			s/([^\r\n]*)[\r\n]+$/$1/ if (!/^ / && !/B\r\n$/);
			&write_sjis($fpwmenu, $_);
		}
		while ($_ = $handle[2]->getline()) {
			&write_sjis($fpwmenu, $_);
			last if (/^\- \- \-/);
		}
		while ($_ = $handle[2]->getline()) {
			s/^ {5}//;
			&write_sjis($fpwmenu, $_)
		}
		$handle[2]->close();
	}

	#
	# construct the index entries.
	#
	&make_index(@list) if (@list);
}

#
# void new_menu_context(char *title, char *tag)
#
sub new_menu_context($$)
{
	my($title, $tag) = @_;

	$fpwmenu->new_context()
		|| die &death_cry("Cannot add a menu context", $fpwmenu);
	if ($fpwmenu->can('add_entry_tag')) {
		$fpwmenu->add_entry_tag($tag)
			|| die &death_cry("Cannot add entry tag: $tag", $fpwmenu);
	} else {
		$fpwmenu->add_tag($tag)
			|| die &death_cry("Cannot add entry tag: $tag", $fpwmenu);
	}
	return if (!$title);

	if (!$fpwmenu->add_keyword_start()
			|| !$fpwmenu->add_text(Jcode::convert($title, 'euc', 'sjis'))
			|| !$fpwmenu->add_keyword_end()
			|| !$fpwmenu->add_newline()
			|| !$fpwmenu->add_indent_level(2)) {
		die &death_cry("Failed to write the menu title", $fpwmenu);
	}
}

#-----------------------------------------------------------------------#
#	WDIC.MAN manipulation												#
#-----------------------------------------------------------------------#
#
# char *write_menu_section(char *title, $handle)
# converts a WDIC.MAN section into a menu entry.
#
sub write_menu_section($$)
{
	my($title, $handle) = @_;
	my($newline, $indent, $prev_indent, $prev_str);

	$newline = 1;
	$indent = $prev_indent = 2;
	$prev_str = '';

	while ($_ = $handle->getline()) {
		if (s/^\-+  (.+)  \-+[\r\n]+$/$1/
				|| s/^\-+[\r\n]+$//) {
			last;

		} elsif (/^[\r\n]+$/) {
			while ($newline < 2) {
				$fpwmenu->add_newline()
					|| die &death_cry("Cannot add newline", $fpwmenu);
				$newline++;
			}
			$prev_str = '';
			next;

		} elsif ('e̓F' eq $title || 'M' eq $title) {
			if (s/^ {5}//) {
				s/^ +/  /;
				if (s/(.*[^ ]  +[^ ].*  +)([^ ].*)$/$1/) {
					$_ .= &reform_sjis($2, 1);
				} elsif (s/  ([^ ].*)$/  /) {
					$_ .= &reform_sjis($1, 1);
				} else {
					$_ = &reform_sjis($_, 1);
				}
				$indent = 3;
			} else {
				s/^ +//;
				s/([^\r\n]+)[\r\n]+$/$1/;
				$indent = 2;
			}

		} elsif ('tH[}bg̑' eq $title) {
			s/^ {3}//;
			if (s/^E//) {
				# E..., heading.
				$indent = 2;
			} elsif (s/^ +(\#)/  $1/
					|| s/^ +( {8}\[)/  $1/
					|| s/^ +(y)/  $1/
					|| s/^ +( {10})/  $1/) {
				# format or yourei example.
				$indent = 3;
			} else {
				s/^ +//;
				s/([^\r\n])[\r\n]+$/$1/;
				$indent = 2;
			}

		} elsif ('iɂ' eq $title) {
			if (!s/^ {3}//) { s/^ {2}//; }
			if (/\x84[\x9f-\xbe]/) {
				# a part of SJIS keisen table.
				$indent = 2;
			} elsif (s/^ {6}([^ ])/  $1/) {
				# imi-hinshi, kakuchou-hinsi.
				$indent = 3;
			} elsif (s/^ {3}([\+\|])/$1/) {
				# imi-hinshi tree.
				if (s/ (E.+)$/ /) { $_ .= &reform_sjis($1, 1); }
				$indent = 3;
			} elsif (s/^ {2}// && (length($_) <= 9 || /^ +\+/)) {
				# kakuchou-hinshi tree.
				s/^ ([^ ])/$1/;
				if (s/ (E.+)$/ /) { $_ .= &reform_sjis($1, 1) };
				$indent = 3;
			} else {
				s/^ +//;
				s/([^\r\n])[\r\n]+$/$1/;
				$indent = 2;
			}

		} elsif ('ǂ݂ɂ' eq $title) {
			s/^ {3}//;
			if (s/^E//) {
				$_ = '' . &reform_sjis($_, 1);
				$indent = 2;
			} elsif (s/^ {3}E//) {
				s/([^\r\n])[\r\n]+$/$1/;
				$indent = 3;
			} elsif (s/^ +(E)/  $1/) {
				$_ = &reform_sjis($_, 1);
				$indent = 3;
			} elsif (s/^ {3}//) {
				if (/^ {14}/ || /^[^ ]+ {2}/) {
					$_ = "  $_";
				} elsif (/^ {2}[^ ]+ {2}/) {
					# nop, french example 'm`etre'.
				} else {
					s/^ +//;
					$_ = &reform_sjis($_, 1) if (!s/([^\.\r\n])[\r\n]+$/$1/);
				}
				$indent = 3;
			} elsif (/^ {2}/) {
				$indent = 3;
			} else {
				s/^ +//;
				$_ = &reform_sjis($_, 1) if (!s/([^\.\r\n])[\r\n]+$/$1/);
				$indent = 2;
			}

		} elsif ('̑̋K' eq $title) {
			if (s/^ E// || s/^ +E//) {
				# the first line or heading.
				$_ = &reform_sjis($_, 1);
				$indent = 2;
			} else {
				s/^ +//;
				if (/ {2}/ || !/[\.\,]/) {
					if (!/  [^ ].*  / && s/  ([^ ].*\x81[\x41\x69].*)$/  /) {
						$_ .= reform_sjis($1, 1);
					}
					$_ = "  $_";
				} else {
					s/([^\r\n])[\r\n]+$/$1/;
				}
				$indent = 3;
			}

		} elsif ('ɂ' eq $title) {
			if (s/^ +E//) {
				$indent = 2;
			} elsif (s/^ {7}([^ ])/$1/) {
				if (/ {2}/) {
					$_ = "  $_";
					$indent = 3;
				} else {
					s/([^\r\n])[\r\n]+$/$1/;
					$indent = 2;
				}
			} elsif (s/^ {7}//) {
				$indent = 3;
			} else {
				s/^ +//;
				s/([^\r\n])[\r\n]+$/$1/;
				$indent = 2;
			}

		} elsif ('b' eq $title) {
			if (s/^ *E([^ \r\n]+[\r\n]+)$/$1/) {
				$_ = '' . &reform_sjis($_, 1);
				$indent = 2;
			} elsif (s/^ {3}E// || /w/) {
				s/^ +//;
				if (s/^(.*[^ ]  +)([^ ].*)$/$2/) {
					$_ = "  $1" . &reform_sjis($_, 1);
				} else {
					$_ = '  ' . &reform_sjis($_, 1);
				}
				$indent = 3;
			} elsif (s/^ {5}( {13})([^ ])/$2/ || s/^ {3}( +)//) {
				$_ = "  $1" . &reform_sjis($_, 1);
			} elsif (s/^ {3}//) {
				if (s/  ([^ ].*)$/  /) {
					$_ .= &reform_sjis($1, 1);
					s/  +(ver)/ $1/i;
				}
				$_ = "  $_";
				$indent = 3;
			} else {
				s/^ +//;
				$indent = 2;
			}

		} elsif ('ₗ' eq $title) {
			s/^ {3}//;
			$indent = 2;
			if (!/^ *[0-9\.]+ \([0-9\/]+\)/
					&& !/^[ \-\=]+[\r\n]+$/
					&& !/ {4}/) {
				if (s/^ *// && !$newline) {
					$fpwmenu->add_newline()
						|| die &death_cry("Cannot add newline", $fpwmenu);
					$newline++;
					$prev_str = '';
				}
				s/^ +//;
				s/([^\r\n]+)[\r\n]+$/$1/;
			}

		} elsif ('T|[g' eq $title) {
			if (s/^ *(\(\x82[\x4f-\x58]\)) */$1 /) {
				$_ = &reform_sjis($_, 1);
				$indent = 2;
			} else {
				s/^ +//;
				s/([^\r\n])[\r\n]+$/$1/ if (!/^[\x20-\x7e]+$/);
				$indent = 3;
			}

		} elsif ('t' eq $title || 'A' eq $title) {
			s/^ +//;
			push(@copyright, $_);

		} elsif (s/^ *E//
				|| s/^ *(\(\%[0-9]\) )/$1/) {
			# a line with bullet or notation index.
			while ($newline < 2) {
				$fpwmenu->add_newline()
					|| die &death_cry("Cannot add newline", $fpwmenu);
				$newline++;
			}
			$prev_str = '';
			s/([^\r\n]*)[\r\n]+$/$1/;
			$indent = 2;

		} elsif (s/^ *([0-9]\. *)/$1 /) {
			# also with an outline number.
			if (!$newline) {
				$fpwmenu->add_newline()
					|| die &death_cry("Cannot add newline", $fpwmenu);
				$newline++;
			}
			$prev_str = '';
			s/([^\r\n]*)[\r\n]+$/$1/;

		} elsif (s/^ *([a-z]+:\/\/[\x21-\x7e]+[\r\n]+)$/  $1/) {
			# URL-only line.
			$indent = 3;

		} else {
			# default processing.
			s/^ +//;
			s/([^\r\n])[\r\n]+$/$1/;
			$indent = 2;
		}

		# write the current string to the menu context.
		if ($indent != $prev_indent) {
			$fpwmenu->add_indent_level($indent)
				|| die &death_cry("Cannot set indent level ($indent)",
					$fpwmenu);
			$prev_indent = $indent;
		}
		$newline = /[\r\n]+$/;
		if ($newline) {
			$prev_str = '';
		} else {
			$_ = &reform_sjis($_, 1);
			if ('' ne $prev_str && $prev_str !~ / $/ && !/^ /) {
				if ($prev_str !~ /$SJIS_ANY$/ || /^[\x21-\x7e]/) {
					$_ = " $_";
				}
			}
			$prev_str = $_;
		}
		&write_sjis($fpwmenu, $_);

		# more fix by my preference.
		if (/^/) {
			while ($newline < 2) {
				$fpwmenu->add_newline()
					|| die &death_cry("Cannot add newline", $fpwmenu);
				$newline++;
			}
		}
	}

	$_ = &reform_sjis($_, 1) if (/\x81[\x69\x6a]/);
	return $_;
}

#-----------------------------------------------------------------------#
#	index construction													#
#-----------------------------------------------------------------------#
#
# void make_index(char **file)
# reads *.TAG files and converts them into index entries.
#
sub make_index(@)
{
	my(@list, %group, $i, $key);

	print "TAG files were found, creating group index...\n";

	@list = $wdic->get_group_list();
	for ($i = 0; $i < scalar(@list); $i++) {
		$group{$list[$i]} = $i;
	}
	@list = sort {
		my $a1 = uc($a);
		my $b1 = uc($b);
		$a1 =~ s/^.*\/([^\/]+)\.TAG/$1/;
		$b1 =~ s/^.*\/([^\/]+)\.TAG/$1/;
		if (defined($group{$a1})) {
			defined($group{$b1}) ? $group{$a1} <=> $group{$b1} : 1;
		} elsif (defined($group{b1})) {
			-1;
		} else {
			$a1 cmp $b1;
		}
	} @_;

	undef %group;
	&new_menu_context('', 'INDEX');

	# write the links to each group.
	foreach (@list) {
		my %hash = (
			HANDLE => new IO::File($_, 'r'),
			TAG => '',
			NAME => ''
		);

		die "Unable to open '$_' ($!)" if (!defined($hash{HANDLE}));
		binmode $hash{HANDLE};

		$_ = uc($_);
		s/^.*\/([^\/]+)\.TAG/$1/;

		$hash{TAG} = $_ . "_TAG";
		$hash{NAME} = $wdic->get_group_title($_);
		$group{$_} = \%hash;

		&write_link($fpwmenu, "$hash{NAME}\n", $hash{TAG});
	}

	# list up the words in the .TAG file.
	foreach $key (keys(%group)) {
		my %hash = %{ $group{$key} };
		print "$key.TAG: $hash{NAME}\n ";

		# create heading index.
		&new_menu_context('', $hash{TAG});
		&write_link($fpwmenu, "($index_title)\n", 'INDEX');

		while ($_ = $hash{HANDLE}->getline()) {
			last if (/^\-+[\r\n]+$/);
			next if (!s/^[A-Z]+\.TAG\([0-9]+\): +y (.+) z.*$/$1/);
			if ('`' eq $_) {
				$_ = 'At@xbg';
			} elsif (/^\x82[\xa0\xa9\xb3\xbd\xc8\xcd\xdc\xe2\xe7]$/) {
				# [Ȃ͂܂]
				$_ .= 's';
			} elsif ('' eq $_) {
				$_ .= 's, ';
			} elsif (length($_) < 4) {
				next;
			}
			&write_link($fpwmenu, "$_\n",
					sprintf("%s%04X", $hash{TAG}, unpack('S', $_)));
		}

		# create word index.
		while ($_ = $hash{HANDLE}->getline()) {
			s/([^\r\n]*)[\r\n]+$/$1/;
			next if (/^\-*$/);

			if (s/^\-+  (.+)  \-+$/$1/) {
				# heading line.
				print $_;
				$newline_on_death = 1;
				if (/^\x82[\x60\xa0\xa9\xb3\xbd\xc8\xcd\xdc\xe2\xe7\xed]$/
						|| 4 <= length($_)) {
					# [`Ȃ͂܂], L, 
					$_ = 'A' if (/^`$/);
					&new_menu_context('',
							sprintf("%s%04X", $hash{TAG}, unpack('S', $_)));
					&write_link($fpwmenu, "($hash{NAME})\n", $hash{TAG});
				}
				next;
			} elsif (!s/^([A-Z]+)\.DIC\([0-9]+\): +\#(.+)$/$2/) {
				next;
			}
			undef @list;
			$list[0] = $1;		# dictionary.

			if (!s/^(.+[^ ]) +(\[.+)$/$2/) {
				die "$key.TAG: Unexpected format: $_";
			}
			$list[1] = $1;		# word.

			if (s/(\]) (.+)$/$1/) {
				# reference to another word.
				$list[2] = $2;	# target.
				&write_sjis($fpwmenu, "$list[1] $_");
				&write_word_link($fpwmenu, " ", $list[2], $key, '');
			} else {
				&write_word_link($fpwmenu, '', $list[1], $key, '');
				&write_sjis($fpwmenu, " $_");
			}
			&write_sjis($fpwmenu,
					' (' . $wdic->get_file_title($list[0]) . ')');
			$fpwmenu->add_newline()
				|| die &death_cry("Cannot add newline", $fpwmenu);
		}

		if ($newline_on_death) {
			print "\n";
			$newline_on_death = 0;
		}
	}
}

#-----------------------------------------------------------------------#
#	writing copyright													#
#-----------------------------------------------------------------------#
#
# void make_copyright(void)
# cretates a copyright entry using WDIC.MAN information and conversion
# date/time.
#
sub make_copyright()
{
	my($str, $sec, $min, $hour, $day, $mon, $year);

	foreach (@copyright) {
		s/([^ ]+) +(.+$)/$2/;
		$str = $1;
		&write_sjis($fpwcopyright, $str . ' ' x (10 - length($str)) . $_);
	}

	$fpwcopyright->add_newline()
		|| die &death_cry("Cannot add newline", $fpwcopyright);

	$str = "This book was generated by ";
	if (defined($ENV{'USER'})) {
		$str .= $ENV{'USER'};
	} elsif (defined($ENV{'USERNAME'})) {
		$str .= $ENV{'USERNAME'};
	} else {
		$str .= '(unknown user)';
	}

	($sec, $min, $hour, $day, $mon, $year) = localtime(time);
	$str .= sprintf(" on %04d/%02d/%02d %02d:%02d:%02d ",
			$year + 1900, $mon, $day, $hour, $min, $sec);

	$VERSION =~ /([0-9\.]+)/;
	$str .= "using fpw-wdic $1.\n";

	$fpwcopyright->add_text($str);
}

1;
